unit MainForm;

interface

uses Windows, Messages, NForm, NCommand, Config;

type
  TDate = record
    Year, Month, Day, YearDay: Integer;
    Holiday: Boolean;
  end;  

  TCalendar = array[0..41] of TDate; 

  TNkCalMainForm = class(TForm)
    Command: TCommand;
    Conf: TConfig;
    Year: Integer;
    Month: Integer;
    OldDay: Integer;
    Today: TSystemTime;
    HolidayColor: COLORREF;
    DayColor: array[0..6] of COLORREF;
    TodayColor: COLORREF;
    OuterDayColor: COLORREF;
    WeekNumberColor: COLORREF;
    BGColor: COLORREF;
    Calendar: TCalendar;
    constructor Create(const Caption: string; ALeft, ATop, AWidth,
      AHeight: Integer); override;
    destructor Destroy; override;
    procedure LoadConfig;
    procedure LoadKeyMap;
    procedure SavePos;
    procedure DrawCalendar(DC: HDC; Year, Month: Integer);
    procedure UpdateDay;
    procedure UpdateCalendar;
    procedure WMContextMenu(var Message: TMessage); message WM_CONTEXTMENU;
    procedure WMCommand(var Message: TWMCommand); message WM_COMMAND;
    procedure WMPaint(var Message: TWMPaint); message WM_PAINT;
    procedure WMTimer(var Message: TWMTimer); message WM_TIMER;
    procedure WMClose(var Message: TWMClose); message WM_CLOSE;
    procedure WMQueryEndSession(var Message: TWMQueryEndSession);
      message WM_QUERYENDSESSION;
  published
    procedure cmAbout;
    procedure cmExit;
    procedure cmNextMonth;
    procedure cmNextYear;
    procedure cmPrevMonth;
    procedure cmPrevYear;
    procedure cmShowWeekNumber;
    procedure cmShowYearDay;
    procedure cmStayOnTop;
    procedure cmThisMonth;
    procedure cmZoom100;
    procedure cmZoom150;
    procedure cmZoom200;
    procedure cmZoom300;
  end;

implementation

uses NLib, NDate, NConf, keymenu, holiday_jp;

{$INCLUDE version.inc}

const
  DayName: array[0..6] of string = (
    'Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat', 'Sun'
  );

function GetCalendar(Year, Month, FirstWeekDay: Integer): TCalendar;
var
  d, i: Integer;
  wd1, wde, py, pm, pdm, prev, ny, nm: Integer;
begin
  FillChar(Result, SizeOf(Result), 0);

  if Month = 0 then
    Exit;

  wd1 := WeekDay(Year, Month, 1);
  wde := WeekDay(Year, Month, DaysInMonth(Year, Month));
  i := 0;
  
  // previous month days
  prev := (wd1 - FirstWeekDay + 7) mod 7;
  if Month = 1 then
  begin
    py := Year-1;
    pm := 12;
  end
  else begin
    py := Year;
    pm := Month-1;
  end;
  pdm := DaysInMonth(py, pm);
  for d := pdm-prev+1 to pdm do
  begin
    Result[i].Year := py;
    Result[i].Month := pm;
    Result[i].Day := d;
    Result[i].YearDay := YearDay(py, pm ,d);
    Inc(i);
  end;

  // this month days
  for d := 1 to DaysInMonth(Year, Month) do
  begin
    Result[i].Year := Year;
    Result[i].Month := Month;
    Result[i].Day := d;
    Result[i].YearDay := YearDay(Year, Month, d);
    Inc(i);
  end;

  // next month days
  if Month = 12 then
  begin
    ny := Year+1;
    nm := 1;
  end
  else begin
    ny := Year;
    nm := Month+1;
  end;
  for d := 1 to (FirstWeekday - wde + 6) mod 7 do
  begin
    Result[i].Year := ny;
    Result[i].Month := nm;
    Result[i].Day := d;
    Result[i].YearDay := YearDay(ny, nm ,d);
    Inc(i);
  end;
end;

constructor TNkCalMainForm.Create(const Caption: string; ALeft, ATop, AWidth,
  AHeight: Integer);
begin
  inherited;
  SetBackground(GetStockObject(NULL_BRUSH));
  Command := TCommand.Create(Self);
  Command.BindKeys(DefaultKeys);
  LoadKeyMap;
  App.Accel := Command.AccelHandle;

  BGColor := RGB(255, 255, 255);
  TodayColor := RGB(230, 234, 242);
  HolidayColor := RGB(255, 0, 0);
  DayColor[0] := RGB(0, 0, 0);
  DayColor[1] := RGB(0, 0, 0);
  DayColor[2] := RGB(0, 0, 0);
  DayColor[3] := RGB(0, 0, 0);
  DayColor[4] := RGB(0, 0, 0);
  DayColor[5] := RGB(72, 72, 240);
  DayColor[6] := RGB(240, 72, 72);
  OuterDayColor := RGB(224, 224, 224);
  WeekNumberColor := RGB(160, 160, 160);

  UpdateDay;
  cmThisMonth;
  SetTimer(Handle, 1, 1000, nil);
  LoadFormPos(JoinPath([App.ProfDir, 'position']), Self);
  LoadConfig;
  UpdateCalendar;
  Show;
end;

destructor TNkCalMainForm.Destroy;
begin
  KillTimer(Handle, 1);
  Command.Free;
  inherited;
end;

procedure TNkCalMainForm.UpdateDay;
begin
  OldDay := Today.wDay;
  GetLocalTime(Today);
  if OldDay <> Today.wDay then 
    UpdateCalendar;
end;

procedure TNkCalMainForm.LoadConfig;
begin
  Config.LoadConfig(JoinPath([App.ProfDir, 'config']), Conf);
end;

procedure TNkCalMainForm.LoadKeyMap;
var
  i: Integer;
  s: array of string;
begin
  SetLength(s, 256);
  for i :=0 to High(s) do
    s[i] := '';
  i := 0;
  try
    with TConf.Create(JoinPath([App.ProfDir, 'keymap'])) do
      try
        while ReadLine and (i < High(s)) do
        begin
          s[i] := GetArg(0);
          s[i+1] := GetArg(1);
          Inc(i, 2);
        end;
      finally
        Free;
      end;
  except
    on E: EIOError do ;
  end;
  SetLength(s, i);
  Command.BindKeys(s);
end;

procedure TNkCalMainForm.SavePos;
begin
  try
    SaveFormPos(JoinPath([App.ProfDir, 'position']), Self);
  except
    on E: EIOError do App.Error(E.Path+EOL+EOL+E.Msg);
  end;
end;

procedure TNkCalMainForm.UpdateCalendar;
begin
  Calendar := GetCalendar(Year, Month, Conf.FirstWDay);
  SetText(Fmt('{0}/{1} - {2}', [Year, Month, App.Name]));
  Invalidate;
end;

procedure DrawText(DC: HDC; Rc: TRect; const S: string);
begin
  Windows.DrawText(DC, PChar(S), Length(S), Rc,
    DT_CENTER or DT_VCENTER or DT_SINGLELINE);
end;

procedure DrawTextCT(DC: HDC; Rc: TRect; const S: string);
begin
  Windows.DrawText(DC, PChar(S), Length(S), Rc,
    DT_CENTER or DT_TOP or DT_SINGLELINE);
end;

procedure DrawTextCB(DC: HDC; Rc: TRect; const S: string);
begin
  Windows.DrawText(DC, PChar(S), Length(S), Rc,
    DT_CENTER or DT_BOTTOM or DT_SINGLELINE);
end;

procedure TNkCalMainForm.DrawCalendar(DC: HDC; Year, Month: Integer);
var
  i, wd, left, x, y, w, h, cw, ch, xx: Integer;
  d: TDate;
  rc: TRect;
  font, mfont, sfont, oldfont: HFONT;
  br: HBRUSH;
  fontsize, fw: Integer;
begin
  GetClientSize(cw, ch);
  if Conf.WeekNumber then
  begin
    w := cw div 8;
    left := w;
  end
  else begin
    w := cw div 7;
    left := 0;
  end;
  h := ch div 8;
  y := 0;

  fontsize := h * 90 div 100;
  if Conf.FontBold then
    fw := FW_BOLD
  else
    fw := FW_NORMAL;
  font := CreateFont(fontsize, 0, 0, 0, fw, Ord(Conf.FontItalic), 0, 0,
    DEFAULT_CHARSET, OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS,
    DEFAULT_QUALITY, DEFAULT_PITCH, PChar(Conf.FontName));
  mfont := CreateFont(fontsize * 3 div 4, 0,0,0, fw, Ord(Conf.FontItalic), 0, 0,
    DEFAULT_CHARSET, OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS,
    DEFAULT_QUALITY, DEFAULT_PITCH, PChar(Conf.FontName));
  sfont := CreateFont(fontsize div 3, 0, 0, 0, fw, Ord(Conf.FontItalic), 0, 0,
    DEFAULT_CHARSET, OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS,
    DEFAULT_QUALITY, DEFAULT_PITCH, PChar(Conf.FontName));

  SetBkMode(DC, TRANSPARENT);
  oldfont := SelectObject(DC, font);

  // Background
  br := CreateSolidBrush(BGColor);
  FillRect(dc, Rect(0, 0, cw, ch), br);
  DeleteObject(br);
  
  // Year and Month
  rc := Rect(0, 0, cw, h);
  if (Year = Today.wYear) and (Month = Today.wMonth) then
  begin
    br := CreateSolidBrush(TodayColor);
    FillRect(DC, rc, br);
    DeleteObject(br);
  end;
  DrawText(DC, rc, Fmt('{0}/{1}', [Year, Month]));
 
  // Day Name
  Inc(y, h);
  SelectObject(DC, mfont);
  if Conf.WeekNumber then
  begin
    SetTextColor(DC, WeekNumberColor);
    DrawText(DC, Rect(0, y, w, y+h), 'Wk');
  end;
  for xx := 0 to 6 do
  begin
    i := (Conf.FirstWDay+xx) mod 7;
    SetTextColor(DC, DayColor[i]);
    DrawText(DC, Rect(left+xx*w, y, left+xx*w+w, y+h), DayName[i]);
  end;

  // Days
  SelectObject(DC, font);
  wd := Conf.FirstWDay;
  for y := 0 to 5 do
  begin
    for x := 0 to 6 do
    begin
      d := Calendar[y*7+x];
      if d.Day = 0 then
        Break;
      
      if Conf.WeekNumber and (wd = 0) then
      begin
        SelectObject(DC, mfont);
        SetTextColor(DC, WeekNumberColor);
        DrawText(DC, Rect(0, y*h+h*2, w, y*h+h*2+h),
          IStr(ISOWeekNumber(d.Year, d.Month, d.Day)));
        SelectObject(DC, font);
      end;

      if d.Month = Month then
      begin
        if IsHoliday(d.Year, d.Month, d.Day) then
          SetTextColor(DC, HolidayColor)
        else
          SetTextColor(DC, DayColor[wd]);
      end
      else if Conf.OuterDays then
        SetTextColor(DC, OuterDayColor)
      else
        SetTextColor(DC, BGColor);

      rc := Rect(left+x*w, y*h+h*2, left+x*w+w, y*h+h*2+h);
      if ((d.Year=Year) and (d.Month=Month) and (d.Day=Today.wDay) and 
           (Year=Today.wYear) and (Month=Today.wMonth)) or
         ((d.Year=Today.wYear) and (d.Month=Today.wMonth) and
           (d.Day=Today.wDay) and Conf.OuterDays) then
      begin
        br := CreateSolidBrush(TodayColor);
        FillRect(DC, rc, br);
        DeleteObject(br);
      end;
      if Conf.YearDay then 
      begin
        DrawTextCT(DC, rc, IStr(d.Day));
        if d.Month = Month then
        begin
          SelectObject(DC, sfont);
          SetTextColor(DC, WeekNumberColor);
          DrawTextCB(DC, rc, Fmt('{0}-{1}', [d.YearDay,
            365+Ord(IsLeap(d.Year))-d.YearDay]));
          SelectObject(DC, font);
        end;
      end else
        DrawText(DC, rc, IStr(d.Day));
      wd := (wd+1) mod 7;
    end;
  end;

  SelectObject(DC, oldfont);
  DeleteObject(font);
  DeleteObject(mfont);
  DeleteObject(sfont);
end; 

{ Command Methods }

procedure TNkCalMainForm.cmStayOnTop;
begin
  SetStayOnTop(not StayOnTop);
end;

procedure TNkCalMainForm.cmPrevMonth;
begin
  if Month = 1 then
  begin
    Month := 12;
    Dec(Year);
  end else
    Dec(Month);
  UpdateCalendar;
end;

procedure TNkCalMainForm.cmNextMonth;
begin
  if Month = 12 then
  begin
     Month := 1;
     Inc(Year);
  end else
     Inc(Month);
  UpdateCalendar;
end;

procedure TNkCalMainForm.cmThisMonth;
begin
  Year := Today.wYear;
  Month := Today.wMonth; 
  UpdateCalendar;
end;

procedure TNkCalMainForm.cmNextYear;
begin
  Inc(Year);
  UpdateCalendar;
end;

procedure TNkCalMainForm.cmPrevYear;
begin
  Dec(Year);
  UpdateCalendar;
end;

procedure TNkCalMainForm.cmZoom100; begin SetClientSize(320, 240) end;
procedure TNkCalMainForm.cmZoom150; begin SetClientSize(480, 360) end;
procedure TNkCalMainForm.cmZoom200; begin SetClientSize(640, 480) end;
procedure TNkCalMainForm.cmZoom300; begin SetClientSize(960, 720) end;

procedure TNkCalMainForm.cmShowWeekNumber;
begin
  Conf.WeekNumber := not Conf.WeekNumber;
  Invalidate;
end;

procedure TNkCalMainForm.cmShowYearDay;
begin
  Conf.YearDay := not Conf.YearDay;
  Invalidate;
end;

procedure TNkCalMainForm.cmAbout;
begin
  App.About(
    App.Name+' '+Version+' ('+ReleaseDate+')'+EOL+
    EOL+
    Copyright+EOL+
    EOL+
    'This software is provided under the zlib/libpng license.'+EOL+
    EOL+
    'Profile Folder: '+App.ProfDir);
end;

procedure TNkCalMainForm.cmExit;
begin
  Close;
end;

{ Message Handlers }

procedure TNkCalMainForm.WMContextMenu(var Message: TMessage);
var
  x, y: Integer;
begin
  ScreenToClient(Message.LParamLo, Message.LParamHi, x, y);
  if y < 0 then
  begin
    inherited;
    Exit;
  end;
  with TPopupMenu.Create do
    try
      Command.AddMenuItems(Handle, MenuItems);
      CheckItem(Command.GetId(cmStayOnTop), StayOnTop);
      CheckItem(Command.GetId(cmShowWeekNumber), Conf.WeekNumber);
      CheckItem(Command.GetId(cmShowYearDay), Conf.YearDay);
      if Message.LParam = -1 then
      begin
        ClientToScreen(0, 0, x, y);
        Popup(Self.Handle, x, y)
      end else
        Popup(Self.Handle, Message.LParamLo, Message.LParamHi);
    finally
      Free;
    end;
end;

procedure TNkCalMainForm.WMTimer(var Message: TWMTimer);
begin
  UpdateDay;
end;

procedure TNkCalMainForm.WMPaint(var Message: TWMPaint);
var
  ps: TPaintStruct;
  dc: HDC;
begin
  dc := BeginPaint(Handle, ps);
  DrawCalendar(dc, Year, Month);
  EndPaint(Handle, ps);
end;

procedure TNkCalMainForm.WMCommand(var Message: TWMCommand);
begin
  Command.DispatchCommand(Message.ItemID);
end;

procedure TNkCalMainForm.WMClose(var Message: TWMClose);
begin
  SavePos;
  inherited;
end;

procedure TNkCalMainForm.WMQueryEndSession(var Message: TWMQueryEndSession);
begin
  SavePos;
  Message.Result := Integer(True);
end;

end.
